/*
  install:
  put innodb_lock_wait.so to @@plugin_dir.
  SQL> INSTALL PLUGIN innodb_lock_wait SONAME 'innodb_lock_wait.so';

  SQL(e.g.) > select * from information_schema.innodb_lock_wait order by ID;

  uninstall:
  SQL> UNINSTALL PLUGIN innodb_lock_wait;
 */

#ifdef USE_PRAGMA_IMPLEMENTATION
#pragma implementation        // gcc: Class implementation
#endif

#define MYSQL_SERVER 1
#include "mysql_priv.h"

#include <mysql/plugin.h>
#include <mysql.h>

#define INSIDE_HA_INNOBASE_CC
#define LOCK_MODULE_IMPLEMENTATION

/* Include necessary InnoDB headers */
extern "C" {
#include "../storage/innobase/include/univ.i"
#include "../storage/innobase/include/os0file.h"
#include "../storage/innobase/include/os0thread.h"
#include "../storage/innobase/include/srv0start.h"
#include "../storage/innobase/include/srv0srv.h"
#include "../storage/innobase/include/trx0roll.h"
#include "../storage/innobase/include/trx0trx.h"
#include "../storage/innobase/include/trx0sys.h"
#include "../storage/innobase/include/mtr0mtr.h"
#include "../storage/innobase/include/row0ins.h"
#include "../storage/innobase/include/row0mysql.h"
#include "../storage/innobase/include/row0sel.h"
#include "../storage/innobase/include/row0upd.h"
#include "../storage/innobase/include/log0log.h"
#include "../storage/innobase/include/lock0lock.h"
#include "../storage/innobase/include/lock0priv.h"
#include "../storage/innobase/include/dict0crea.h"
#include "../storage/innobase/include/btr0cur.h"
#include "../storage/innobase/include/btr0btr.h"
#include "../storage/innobase/include/fsp0fsp.h"
#include "../storage/innobase/include/sync0sync.h"
#include "../storage/innobase/include/fil0fil.h"
#include "../storage/innobase/include/trx0xa.h"
#include "../storage/innobase/include/thr0loc.h"
#include "../storage/innobase/include/ha_prototypes.h"

/*
  Wrappers for inline functions
 */

void
mutex_enter_wrap(mutex_t* mutex);

trx_t*
ut_list_get_first_wrap_trx();

lock_t*
lock_rec_get_first_on_page_addr_wrap(ulint space, ulint page_no);

ibool
lock_rec_get_nth_bit_wrap(lock_t* lock, ulint i);

ulint
ut_dulint_get_high_wrap(dulint d);

ulint
ut_dulint_get_low_wrap(dulint d);

lock_t*
lock_rec_get_next_on_page_wrap(lock_t* lock);

lock_t*
ut_list_get_first_wrap_locks(dict_table_t* table);

lock_t*
ut_list_get_next_wrap_locks(lock_t* lock);

trx_t*
ut_list_get_next_wrap_trx(trx_t* trx);

void
mutex_exit_wrap(mutex_t* mutex);

}

#define LOCK_MODE_STR_LENGTH 4

static void innodb_row_lock_conv_mode(lock_t* lock, char* buff)
{
    switch(lock->type_mode & LOCK_MODE_MASK)
    {
	case LOCK_S:
	    buff[0]='S';
	    break;
	case LOCK_X:
	    buff[0]='X';
	    break;
	case LOCK_NONE:
	    buff[0]='N';
	    break;
	default:
	    buff[0]='?';
    }
    switch(lock->type_mode & (LOCK_GAP | LOCK_REC_NOT_GAP))
    {
	case LOCK_ORDINARY:
	    buff[1]='O';
	    break;
	case LOCK_GAP:
	    buff[1]='G';
	    break;
	case LOCK_REC_NOT_GAP:
	    buff[1]='-';
	    break;
	default:
	    buff[1]='?';
    }
    buff[2]=(lock->type_mode & LOCK_INSERT_INTENTION)?'I':'-';
    buff[3]=(lock->type_mode & LOCK_WAIT)?'W':'-';
    buff[LOCK_MODE_STR_LENGTH]=0;
}


static void innodb_table_lock_conv_mode(lock_t* lock, char* buff)
{
    switch(lock->type_mode & LOCK_MODE_MASK)
    {
	case LOCK_S:
	    buff[0]=' '; buff[1]='S';
	    break;
	case LOCK_X:
	    buff[0]=' '; buff[1]='X';
	    break;
	case LOCK_IS:
	    buff[0]='I'; buff[1]='S';
	    break;
	case LOCK_IX:
	    buff[0]='I'; buff[1]='X';
	    break;
	case LOCK_AUTO_INC:
	    buff[0]='A'; buff[1]='I';
	    break;
	default:
	    buff[0]='?'; buff[1]='?';
	    break;
    }
    buff[2]='-';
    buff[3]=(lock->type_mode & LOCK_WAIT)?'W':'-';
    buff[LOCK_MODE_STR_LENGTH]=0;
}


static int innodb_lock_wait_fill_table(THD *thd, TABLE_LIST *tables, COND *cond)
{
    TABLE *table= (TABLE *) tables->table;

    trx_t* trx;
    lock_t* lock;
    dict_table_t* dict_table;
    ulint space, page_no, heap_no;
    char buff[LOCK_MODE_STR_LENGTH + 1];

    DBUG_ENTER("innodb_lock_wait_fill_table");
    int returnable= 0;

    pthread_mutex_lock(&LOCK_thread_count);
    mutex_enter_wrap(&kernel_mutex);

    trx = ut_list_get_first_wrap_trx();

    while(trx)
    {
	if(trx->wait_lock)
	{
	    if((trx->wait_lock->type_mode & LOCK_TYPE_MASK) == LOCK_REC)
	    {
		space = trx->wait_lock->un_member.rec_lock.space;
		page_no = trx->wait_lock->un_member.rec_lock.page_no;
		heap_no = lock_rec_find_set_bit(trx->wait_lock);

		lock = lock_rec_get_first_on_page_addr_wrap(space, page_no);

		while (lock != trx->wait_lock) {

		    if (lock_rec_get_nth_bit_wrap(lock, heap_no)
			&& lock_has_to_wait(trx->wait_lock, lock)) {

			table->field[0]->store(((THD*)(trx->mysql_thd))->thread_id);
			table->field[1]->store(ut_dulint_get_high_wrap(trx->id));
			table->field[2]->store(ut_dulint_get_low_wrap(trx->id));
			table->field[3]->store(difftime(time(NULL), trx->wait_started));
			table->field[4]->store(trx->wait_lock->index->table_name,
					       strlen(trx->wait_lock->index->table_name),
					       system_charset_info);
			table->field[5]->store(trx->wait_lock->index->name,
					       strlen(trx->wait_lock->index->name),
					       system_charset_info);
			table->field[6]->store(space);
			table->field[7]->store(page_no);
			table->field[8]->store(lock_rec_find_set_bit(trx->wait_lock));

			innodb_row_lock_conv_mode(trx->wait_lock, buff);
			table->field[9]->store(buff,LOCK_MODE_STR_LENGTH,system_charset_info);

			table->field[10]->store(((THD*)(lock->trx->mysql_thd))->thread_id);
			table->field[11]->store(ut_dulint_get_high_wrap(lock->trx->id));
			table->field[12]->store(ut_dulint_get_low_wrap(lock->trx->id));

			innodb_row_lock_conv_mode(lock, buff);
			table->field[13]->store(buff,LOCK_MODE_STR_LENGTH,system_charset_info);

			if (schema_table_store_record(thd, table))
			{
			    returnable= 1;
			    goto end_func;
			}

		    }
		    lock = lock_rec_get_next_on_page_wrap(lock);
		}
	    }
	    else if((trx->wait_lock->type_mode & LOCK_TYPE_MASK) == LOCK_TABLE)
	    {
		dict_table = trx->wait_lock->un_member.tab_lock.table;

		lock = ut_list_get_first_wrap_locks(dict_table);

		while (lock != trx->wait_lock) {

		    if(lock_has_to_wait(trx->wait_lock, lock)) {

			table->field[0]->store(((THD*)(trx->mysql_thd))->thread_id);
			table->field[1]->store(ut_dulint_get_high_wrap(trx->id));
			table->field[2]->store(ut_dulint_get_low_wrap(trx->id));
			table->field[3]->store(difftime(time(NULL), trx->wait_started));
			table->field[4]->store(dict_table->name,
					       strlen(dict_table->name),
					       system_charset_info);
			table->field[5]->store("",
					       0,
					       system_charset_info);
			table->field[6]->store(0);
			table->field[7]->store(0);
			table->field[8]->store(0);

			innodb_table_lock_conv_mode(trx->wait_lock, buff);
			table->field[9]->store(buff,LOCK_MODE_STR_LENGTH,system_charset_info);

			table->field[10]->store(((THD*)(lock->trx->mysql_thd))->thread_id);
			table->field[11]->store(ut_dulint_get_high_wrap(lock->trx->id));
			table->field[12]->store(ut_dulint_get_low_wrap(lock->trx->id));

			innodb_table_lock_conv_mode(lock, buff);
			table->field[13]->store(buff,LOCK_MODE_STR_LENGTH,system_charset_info);

			if (schema_table_store_record(thd, table))
			{
			    returnable= 1;
			    goto end_func;
			}
		    }
		    lock = ut_list_get_next_wrap_locks(lock);
		}
	    }
	}

	trx = ut_list_get_next_wrap_trx(trx);
    }

 end_func:
    mutex_exit_wrap(&kernel_mutex);
    pthread_mutex_unlock(&LOCK_thread_count);

    DBUG_RETURN(returnable);
}

ST_FIELD_INFO innodb_lock_wait_field_info[]=
{
    {"ID", 4, MYSQL_TYPE_LONG, 0, 0, "Thread id", SKIP_OPEN_TABLE},
    {"TRX_H", 4, MYSQL_TYPE_LONG, 0, 0, "TRX waiting", SKIP_OPEN_TABLE},
    {"TRX_L", 4, MYSQL_TYPE_LONG, 0, 0, "TRX waiting", SKIP_OPEN_TABLE},
    {"SECS", 4, MYSQL_TYPE_LONG, 0, 0, "Waiting period", SKIP_OPEN_TABLE},
    {"TABLE", 256, MYSQL_TYPE_STRING, 0, 0, "Table_name", SKIP_OPEN_TABLE},
    {"INDEX", 256, MYSQL_TYPE_STRING, 0, 0, "Index_name", SKIP_OPEN_TABLE},
    {"SP", 4, MYSQL_TYPE_LONG, 0, 0, "Space No.", SKIP_OPEN_TABLE},
    {"PG", 4, MYSQL_TYPE_LONG, 0, 0, "Page No.", SKIP_OPEN_TABLE},
    {"BT", 4, MYSQL_TYPE_LONG, 0, 0, "Bit No.", SKIP_OPEN_TABLE},
    {"MODE", 5, MYSQL_TYPE_STRING, 0, 0, "Type_mode", SKIP_OPEN_TABLE},
    {"WID", 4, MYSQL_TYPE_LONG, 0, 0, "Thread id waited", SKIP_OPEN_TABLE},
    {"WTRX_H", 4, MYSQL_TYPE_LONG, 0, 0, "TRX waited", SKIP_OPEN_TABLE},
    {"WTRX_L", 4, MYSQL_TYPE_LONG, 0, 0, "TRX waited", SKIP_OPEN_TABLE},
    {"WMODE", 5, MYSQL_TYPE_STRING, 0, 0, "Type_mode", SKIP_OPEN_TABLE},

    {0, 0, MYSQL_TYPE_STRING, 0, 0, 0, SKIP_OPEN_TABLE}
};

static int innodb_lock_wait_init(void *p)
{
  DBUG_ENTER("innodb_lock_wait_init");
  ST_SCHEMA_TABLE *schema= (ST_SCHEMA_TABLE *)p;

  schema->fields_info= innodb_lock_wait_field_info;
  schema->fill_table= innodb_lock_wait_fill_table;

  DBUG_RETURN(0);
}

static int innodb_lock_wait_deinit(void *p)
{
  DBUG_ENTER("innodb_lock_wait_deinit");
  DBUG_RETURN(0);
}

struct st_mysql_information_schema innodb_lock_wait=
{ MYSQL_INFORMATION_SCHEMA_INTERFACE_VERSION };

mysql_declare_plugin(innodb_lock_wait)
{
  MYSQL_INFORMATION_SCHEMA_PLUGIN,
  &innodb_lock_wait,
  "innodb_lock_wait",
  "Yasufumi Kinoshita",
  "innodb locks being waited other innodb locks.(experimental!)",
  PLUGIN_LICENSE_BSD,
  innodb_lock_wait_init, /* Plugin Init */
  innodb_lock_wait_deinit, /* Plugin Deinit */
  0x0003,
  NULL,                       /* status variables                */
  NULL,          /* system variables */
  NULL                        /* config options                  */
}
mysql_declare_plugin_end;
